/**
 * Mozilla Archive Format
 * ======================
 *
 * Version: 0.2.17
 *
 * Author: Christopher Ottley
 *
 * Description: First public release of the MAF extension for Firefox and Mozilla
 *
 *  Copyright (c) 2004 Christopher Ottley.
 *
 *  This file is part of MAF.
 *
 *  MAF is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  MAF is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with MAF; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * TODO:
 *   Implement open archive functionality
 *     - Based on preferences
 *       - Dialog box showing all archived files, select to open - DONE
 *       - Open all in new tabs - DONE
 *     - URL Rewrite preference
 *       - As tabs are loaded, search for URLs that match archived URLs and replace with local file:// URLs.
 *     - Show loading dialog until archives have decompressed - DONE
 *
 *   Implement save archive functionality
 *     - Save additional meta data (search for METAXXX).
 *
 *   Implement Preferences functionality
 *     - Settings to create the script used in archiving (Should I?)
 *     - Whole Prefences GUI and logic insanity
 *       - Browse button for existing files should change to the directory (if possible)
 *
 *   Optimizations:
 *     - Look at implementing an Observer object for the download instead of a timer on a 1 second delay.
 *     - Look also for javascript observer for the save tabs dialog
 *
 *   Additional Planned Features:
 *     - MHT File support via a pseudo-app that is responsible for MHT decoding and encoding
 *
 *   Known Bugs:
 *     - Other content types (like PDF) haven't been tested
 *     - The GUI does not have the ability to add more archive app scripts or extensions
 *     - If there is a : in the title of something, the RDF reader freaks out
 */


/**
 * Contracts and interfaces used in the source.
 */
const localFileContractID = "@mozilla.org/file/local;1";
const localFileIID = Components.interfaces.nsILocalFile;
const processContractID = "@mozilla.org/process/util;1";
const processIID = Components.interfaces.nsIProcess;
const filePickerContractID = "@mozilla.org/filepicker;1";
const filePickerIID = Components.interfaces.nsIFilePicker;
const fileOutputStreamContractID = "@mozilla.org/network/file-output-stream;1";
const fileOutputStreamIID = Components.interfaces.nsIFileOutputStream;
const externalHelperAppServiceContractID = "@mozilla.org/uriloader/external-helper-app-service;1";
const externalHelperAppServiceIID = Components.interfaces.nsIExternalHelperAppService;
const downloadContractID = "@mozilla.org/download;1";
const downloadIID = Components.interfaces.nsIDownload;
const ioServiceNetworkContractID = "@mozilla.org/network/io-service;1";
const ioServiceNetworkIID = Components.interfaces.nsIIOService;
const xmlRDFDatasourceContractID = "@mozilla.org/rdf/datasource;1?name=xml-datasource";
const xmlRDFDatasourceIID = Components.interfaces.nsIRDFRemoteDataSource;
const rdfRDFServiceContractID = "@mozilla.org/rdf/rdf-service;1";
const rdfRDFServiceIID = Components.interfaces.nsIRDFService;
const prefSvcContractID = "@mozilla.org/preferences-service;1";
const prefSvcIID = Components.interfaces.nsIPrefService;
const rdfRDFCServiceContractID = "@mozilla.org/rdf/container-utils;1";
const rdfRDFCServiceIID = Components.interfaces.nsIRDFContainerUtils;
const appShellContractID = "@mozilla.org/appshell/appShellService;1";
const appShellIID = Components.interfaces.nsIAppShellService;
const asWinMedContractID = "@mozilla.org/appshell/window-mediator;1";
const asWinMedIID = Components.interfaces.nsIWindowMediator;

const urlIID = Components.interfaces.nsIURL;
const webBrowserPersistIID = Components.interfaces.nsIWebBrowserPersist;
const rdfDatasourceIID = Components.interfaces.nsIRDFDataSource;

const MAFNamespaceId = "MAF";
const MAFNamespace = "http://maf.mozdev.org/metadata/rdf#";

const MAFRDFTemplate = '<?xml version="1.0"?>\n' +
  '<RDF:RDF xmlns:'+ MAFNamespaceId +'="'+ MAFNamespace +'"\n' +
  '     xmlns:NC="http://home.netscape.com/NC-rdf#"\n' +
  '     xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n' +
  '  <RDF:Description about="urn:root">\n'+
  '  </RDF:Description>\n' +
  '</RDF:RDF>\n';

/** The global RDF service object. */
var gRDFService = Components.classes[rdfRDFServiceContractID].getService(rdfRDFServiceIID);

/** The global RDF Container Service object. */
var gRDFCService = Components.classes[rdfRDFCServiceContractID].getService(rdfRDFCServiceIID);


/**
 * Saving the document specified occurs in a seperate thread,
 * so an object with its own timer is used to check the status
 * of the download before an archive script is invoked.
 */
function MafArchiver(aDocument, tempPath, scriptPath, archivePath, dateTimeArchived) {
  /** The document to archive. */
  this.aDocument = aDocument,

  /** The path of the temp folder to use. */
  this.tempPath = tempPath,

  /** The path of the archive script to use. */
  this.scriptPath = scriptPath,

  /** The full path of the archive file to archive to. */
  this.archivePath = archivePath,

  /** When this document was archived. */
  this.dateTimeArchived = dateTimeArchived,

  /** The folder number used in the archive. */
  this.folderNumber = dateTimeArchived.valueOf()+"_"+Math.floor(Math.random()*1000),

  /** Keeps track of the timer used to check the download status. */
  this.timerId = null,

  /** Flag to ensure the start function isn't called twice. */
  this.started = false,

  this.downloadComplete = false,

  /**
   * Starts the download process and the interval checking the download status.
   */
  this.start = function() {
    if (!this.started) {
      this.started = true;
      this._ns_sniffHeader(this.aDocument, MafUtils.appendToDir(this.tempPath, this.folderNumber), "index.html");
      this.timerId = setInterval(this._checkDownloadComplete, 1000, this);
    }
  },

  /*
  this.downloadProgressListener = {
    onStateChange: function(webProgress, request, stateFlags, status) { },
    onProgressChange: function(webProgress, request, curSelfProgress,
                                maxSelfProgress, curTotalProgress, maxTotalProgress) {

          if (maxTotalProgress == -1) {
            this._checkDownloadComplete(this);
          }
          alert("onProgressChange, curSelfProgress: " + curSelfProgress +
                       " maxSelfProgress: " + maxSelfProgress +
                       " curTotalProgress: " + curTotalProgress +
                       " maxTotalProgress: " + maxTotalProgress);
    },
    onStatusChange: function(webProgress, request, status, message) { },
    onSecurityChange: function(webProgress, request, state) { },
    QueryInterface: function(iid) {
      if(!iid.equals(Components.interfaces.nsIWebProgressListener) &&
         !iid.equals(Components.interfaces.nsISupportsWeakReference)) {
           throw Components.results.NS_ERROR_NO_INTERFACE;
      }
      return this;
    }
  }, */

  /**
   * Clears the interval timer if the download is complete.
   * Also triggers the archive download function.
   */
  this._checkDownloadComplete = function(objMafArchiver) {
    // If the download is complete, add to the archive
    if (objMafArchiver.dl!=null) {
      if (objMafArchiver.dl.percentComplete == 100) {
        clearInterval(objMafArchiver.timerId);
        objMafArchiver.dl = null;
        objMafArchiver.downloadComplete = true;
        objMafArchiver.addMetaData(objMafArchiver);
        objMafArchiver.archiveDownload(objMafArchiver.scriptPath,
                                       objMafArchiver.archivePath,
                                       objMafArchiver.folderNumber);

        // Remove temporary directory after 5 seconds
        setTimeout(this._removeFolder, 5000, MafUtils.appendToDir(objMafArchiver.tempPath, objMafArchiver.folderNumber));
      }
    }
  },

  /**
   * Removes the specified folder if it exists.
   */
  this._removeFolder = function(folderToRemove) {
    var oDir = Components.classes[localFileContractID].getService(localFileIID);
    oDir.initWithPath(folderToRemove);
    if (oDir.exists() && oDir.isDirectory()) {
      oDir.remove(true);
    }
  },

  /**
   * Adds RDF files containing Meta Data about the saved page
   * such as original URL, date/time saved and page history
   * METAXXX - Page history not implemented as yet.
   * METAXXX - For future, also save scroll position and text zoom.
   * METAXXX - History index
   */
  this.addMetaData = function(objMafArchiver) {

    destMetaDataFolder = MafUtils.appendToDir(objMafArchiver.tempPath, objMafArchiver.folderNumber);
    // Create index.rdf in the folderNumber and
    // Get a referance to index.rdf's data source
    var indexDS = MafUtils.createRDF(destMetaDataFolder, "index.rdf");

    // Add url data
    MafUtils.addStringData(indexDS, "originalurl", objMafArchiver.aDocument.location.href);
    // Add title
    MafUtils.addStringData(indexDS, "title", objMafArchiver.aDocument.title);
    // Add Date/Time archived data
    MafUtils.addStringData(indexDS, "archivetime", objMafArchiver.dateTimeArchived);
    // Add index file data
    MafUtils.addStringData(indexDS, "indexfilename", objMafArchiver.indexfilename);

    // Write changes to physical file
    indexDS.Flush();


    // Create history.rdf in the folderNumber and
    // Get a reference to history.rdf
    //var historyDS = MafUtils.createRDF(destMetaDataFolder, "history.rdf");

    // Add history information

    // Write changes to physical file
    //historyDS.Flush();
  },

  /**
   * Archives the downloaded file(s).
   */
  this.archiveDownload = function(program, archivefile, sourcepath) {
    /** If program is nothing then don't try to run it. */
    if (program != "") {

      if (MafPreferences.win_invisible) {
        localProgram = MafPreferences.win_wscriptexe;
        localProgramArgs = new Array();
        localProgramArgs[localProgramArgs.length] = MafPreferences.win_invisiblevbs;
        localProgramArgs[localProgramArgs.length] = program;
      } else {
        localProgram = program;
        localProgramArgs = new Array();
      }

      try {
        var oProgram = Components.classes[localFileContractID].getService(localFileIID);
        oProgram.initWithPath(localProgram);
      } catch(e) {
        alert("Could not find program: " + program + " \n" + e);
      }

      try {
        var oProcess = Components.classes[processContractID].createInstance(processIID);
      } catch (e) {
        alert("Could not create process:\n" + e);
      }

      oProcess.init(oProgram);

      localProgramArgs[localProgramArgs.length] = archivefile;
      localProgramArgs[localProgramArgs.length] = sourcepath;

      oProcess.run(true, localProgramArgs, localProgramArgs.length);
    }
  },

  /**
   * Trigger the file save process with the specified data structure.
   */
  this._ns_sniffHeader = function(aDocument, aSaveDocPath, aSaveDocFileName) {
    var data = {
      url: aDocument.location.href,
      fileName: null,
      filePickerTitle: null,
      document: aDocument,
      bypassCache: false,
      window: window,
      saveDocPath: aSaveDocPath,
      saveDocFileName: aSaveDocFileName,
      objMafArchiver: this
    };

    /** Change the header sniffer callback to use the this._ns_foundHeaderInfo function instead */
    var sniffer = new nsHeaderSniffer(aDocument.location.href, this._ns_foundHeaderInfo, data, true);
  },

  /**
   * Based on code found in foundHeaderInfo in contentAreaUtils.js
   * except this._ns_getTargetFile is called instead.
   * Changed contract ids and interface references to use global constants.
   */
  this._ns_foundHeaderInfo = function(aSniffer, aData, aSkipPrompt) {
    var contentType = aSniffer.contentType;
    var contentEncodingType = aSniffer.contentEncodingType;

    var shouldDecode = false;
    // Are we allowed to decode?
    try {
      const helperAppService =
        Components.classes[externalHelperAppServiceContractID].getService(externalHelperAppServiceIID);
      var url = aSniffer.uri.QueryInterface(urlIID);
      var urlExt = url.fileExtension;
      if (helperAppService.applyDecodingForExtension(urlExt,
                                                   contentEncodingType)) {
        shouldDecode = true;
      }
    }
    catch (e) {
    }

    var isDocument = aData.document != null && isDocumentType(contentType);
    if (!isDocument && !shouldDecode && contentEncodingType) {
      // The data is encoded, we are not going to decode it, and this is not a
      // document save so we won't be doing a "save as, complete" (which would
      // break if we reset the type here).  So just set our content type to
      // correspond to the outermost encoding so we get extensions and the like
      // right.
      contentType = contentEncodingType;
    }

    var file = null;
    var saveAsType = kSaveAsType_URL;
    try {
      file = aData.fileName.QueryInterface(localFileIID);
    }
    catch (e) {
      var saveAsTypeResult = { rv: 0 };
      file = aData.objMafArchiver._ns_getTargetFile(aData, aSniffer, contentType, isDocument, aSkipPrompt, saveAsTypeResult);
      saveAsType = saveAsTypeResult.rv;
    }

    // If we're saving a document, and are saving either in complete mode or
    // as converted text, pass the document to the web browser persist component.
    // If we're just saving the HTML (second option in the list), send only the URI.
    var source = (isDocument && saveAsType != kSaveAsType_URL) ? aData.document : aSniffer.uri;
    var persistArgs = {
      source      : source,
      contentType : (isDocument && saveAsType == kSaveAsType_Text) ? "text/plain" : contentType,
      target      : file,
      postData    : aData.document ? getPostData() : null,
      bypassCache : aData.bypassCache
    };

    var persist = makeWebBrowserPersist();

    // Calculate persist flags.
    const flags = webBrowserPersistIID.PERSIST_FLAGS_NO_CONVERSION | webBrowserPersistIID.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
    if (aData.bypassCache)
      persist.persistFlags = flags | webBrowserPersistIID.PERSIST_FLAGS_BYPASS_CACHE;
    else
      persist.persistFlags = flags | webBrowserPersistIID.PERSIST_FLAGS_FROM_CACHE;

    if (shouldDecode)
      persist.persistFlags &= ~webBrowserPersistIID.PERSIST_FLAGS_NO_CONVERSION;

    // Create download and initiate it (below)
    aData.objMafArchiver.dl = Components.classes[downloadContractID].createInstance(downloadIID);

    try {

    if (isDocument && saveAsType != kSaveAsType_URL) {
      // Saving a Document, not a URI:
      var filesFolder = null;
      if (persistArgs.contentType != "text/plain") {
        // Create the local directory into which to save associated files.
        filesFolder = Components.classes[localFileContractID].createInstance(localFileIID);
        filesFolder.initWithPath(persistArgs.target.path);

        var nameWithoutExtension = filesFolder.leafName;
        nameWithoutExtension = nameWithoutExtension.substring(0, nameWithoutExtension.lastIndexOf("."));
        var filesFolderLeafName = getStringBundle().formatStringFromName("filesFolder",
                                                                       [nameWithoutExtension],
                                                                       1);

        filesFolder.leafName = filesFolderLeafName;
      }

      var encodingFlags = 0;
      if (persistArgs.contentType == "text/plain") {
        encodingFlags |= webBrowserPersistIID.ENCODE_FLAGS_FORMATTED;
        encodingFlags |= webBrowserPersistIID.ENCODE_FLAGS_ABSOLUTE_LINKS;
        encodingFlags |= webBrowserPersistIID.ENCODE_FLAGS_NOFRAMES_CONTENT;
      }

      const kWrapColumn = 80;
      aData.objMafArchiver.dl.init(aSniffer.uri, persistArgs.target, null, null, null, persist);
      // The progress listener isn't working!
      // The progress finishes but the download window comes up and says "starting..."
      //var dlListener = aData.objMafArchiver.downloadProgressListener;
      //dlListener = dlListener.QueryInterface(Components.interfaces.nsIWebProgressListener);
      //persist.progressListener = dlListener;
      persist.saveDocument(persistArgs.source, persistArgs.target, filesFolder,
                         persistArgs.contentType, encodingFlags, kWrapColumn);
    } else {
      aData.objMafArchiver.dl.init(source, persistArgs.target, null, null, null, persist);
      // The progress listener isn't working!
      // The progress finishes but the download window comes up and says "starting..."
      //var dlListener = aData.objMafArchiver.downloadProgressListener;
      //dlListener = dlListener.QueryInterface(Components.interfaces.nsIWebProgressListener);
      //persist.progressListener = dlListener;
      persist.saveURI(source, null, null, persistArgs.postData, null, persistArgs.target);
    }

    } catch(e) {

    }


  },

  /**
   * Based on code found in getTargetFile in contentAreaUtils.js
   * except it does not prompt the user for the file name and path.
   * Instead uses values stored in the aData record structure.
   */
  this._ns_getTargetFile = function(aData, aSniffer, aContentType, aIsDocument, aSkipPrompt, aSaveAsTypeResult) {
  aSaveAsTypeResult.rv = kSaveAsType_Complete;

  // Determine what the 'default' string to display in the File Picker dialog
  // should be.
  var defaultFileName = getDefaultFileName(aData.fileName,
                                           aSniffer.suggestedFileName,
                                           aSniffer.uri,
                                           aData.document);

  var defaultExtension = getDefaultExtension(defaultFileName, aSniffer.uri, aContentType);

  var defaultString = getNormalizedLeafName("index", defaultExtension);

  aData.objMafArchiver.indexfilename = defaultString;

    /** Used to be the default download folder preference. Now it's whatever folder was specified in the
        data structure. */
  var dir = null;
  try {
    dir = Components.classes[localFileContractID].getService(localFileIID);
    dir.initWithPath(aData.saveDocPath);

    // Make the directory!!!
    if (!dir.exists()) {
      dir.create(0x01, 0777);
    }
  }
  catch (e) {

  }

  var file;

    // ben 07/31/2003:
    // We don't nullcheck dir here because dir should never be null if we get here
    // unless something is badly wrong, and if it is, I want to know about it in
    // bugs.
    dir.append(defaultString);
    file = dir;

    // Since we're automatically downloading, we don't get the file picker's
    // logic to check for existing files, so we need to do that here.
    //
    // Note - this code is identical to that in
    //   browser/components/downloads/content/nsHelperAppDlg.js.
    // If you are updating this code, update that code too! We can't share code
    // here since that code is called in a js component.
    while (file.exists()) {
      var parts = /.+-(\d+)(\..*)?$/.exec(file.leafName);
      if (parts) {
        file.leafName = file.leafName.replace(/((\d+)\.)/,
                                              function (str, p1, part, s) {
                                                return (parseInt(part) + 1) + ".";
                                              });
      }
      else {
        file.leafName = file.leafName.replace(/\./, "-1$&");
      }
    }


  return file;

  }


};

/**
 * Allows the archiving of multiple tabs.
 */
function MafTabArchiver(browsers, tempPath, scriptPath, archivePath) {
  /** The tabs to archive. */
  this.browsers = browsers,

  /** The path of the temp folder to use. */
  this.tempPath = tempPath,

  /** The path of the archive script to use. */
  this.scriptPath = scriptPath,

  /** The full path of the archive file to archive to. */
  this.archivePath = archivePath,

  /** Keeps track of the timer used to check the download status. */
  this.timerId = null,

  /** Flag to ensure the start function isn't called twice. */
  this.started = false,

  /** If all the tabs have finished downloading. */
  this.downloadComplete = false,

  /** The archiver objects used to save. */
  this.MafArchivers = new Array(),

  /** The current tab being saved. */
  this.currentMafArchiverIndex = 0;

  /**
   * Starts the download process and the interval checking the download status.
   * An object is created for each tab that is saving,
   * when it is finished, the next one starts.
   */
  this.start = function() {
    if (!this.started) {
      this.started = true;
      // Should always be true, but eh.
      if (browsers.length>0) {
        this.MafArchivers[this.MafArchivers.length] = new MafArchiver(browsers[0].contentDocument, tempPath, scriptPath, archivePath, new Date());
        this.MafArchivers[this.currentMafArchiverIndex].start();
        this.timerId = setInterval(this._checkDownloadComplete, 1000, this);
      }
    }
  },


  this.stop = function() {
    clearInterval(this.timerId);
  }

  /**
   * Clears the interval timer if the download is complete.
   * Also triggers the archive download function.
   */
  this._checkDownloadComplete = function(objMafTabArchiver) {
    // If the current download is complete, start the next one

    if (objMafTabArchiver.MafArchivers!=null) {
      if (objMafTabArchiver.currentMafArchiverIndex < objMafTabArchiver.browsers.length) {
        if (objMafTabArchiver.MafArchivers[objMafTabArchiver.currentMafArchiverIndex].downloadComplete == true) {
          if (objMafTabArchiver.fnProgressUpdater != null) {
            percentage = Math.floor((objMafTabArchiver.currentMafArchiverIndex/objMafTabArchiver.browsers.length)*100);
            objMafTabArchiver.fnProgressUpdater(percentage);
          }
          objMafTabArchiver.currentMafArchiverIndex += 1;
          objMafTabArchiver.MafArchivers[objMafTabArchiver.MafArchivers.length] = new MafArchiver(objMafTabArchiver.browsers[objMafTabArchiver.currentMafArchiverIndex].contentDocument, tempPath, scriptPath, archivePath, new Date());
          objMafTabArchiver.MafArchivers[objMafTabArchiver.currentMafArchiverIndex].start();
        }
      } else {
        clearInterval(objMafTabArchiver.timerId);
        if (objMafTabArchiver.fnProgressUpdater != null) {
          objMafTabArchiver.fnProgressUpdater(100);
        }
        objMafTabArchiver.downloadComplete = true;
      }
    } else {
      clearInterval(objMafTabArchiver.timerId);
    }
  },

  /**
   * Access method to set the progress update function
   * This is the function that is called whenever there is a change
   * in progress
   */
  this.setProgressUpdater = function(fnProgressUpdater) {
    this.fnProgressUpdater = fnProgressUpdater;
  }

};


/**
 * Extracts the archive to the specified path
 */
function MafTabExpander(tempPath, scriptPath, archivePath, folderNumber) {
  /** The path of the temp folder to use. */
  this.tempPath = tempPath,

  /** The path of the archive script to use. */
  this.scriptPath = scriptPath,

  /** The full path of the archive file to archive to. */
  this.archivePath = archivePath,

  /** Flag to ensure the start function isn't called twice. */
  this.started = false,

  /** The folder holding the expanded archive contents. */
  this.folderNumber = folderNumber,

  /**
   * Schedules the extract process to start.
   */
  this.start = function() {
    if (!this.started) {
      this.started = true;
      this.timerId = setTimeout(this._startExtract, 500, this);
    }
  },

  /**
   * Starts the extract process.
   */
  this._startExtract = function(objExtractor) {
    Maf.extractFromArchive(objExtractor.scriptPath, objExtractor.archivePath, MafUtils.appendToDir(objExtractor.tempPath, objExtractor.folderNumber));
    objExtractor.fnProgressUpdater(100);
  },


  /**
   * Can't kill the process, so stop does nothing really.
   */
  this.stop = function() {

  },

  /**
   * Access method to set the progress update function
   * This is the function that is called whenever there is a change
   * in progress
   */
  this.setProgressUpdater = function(fnProgressUpdater) {
    this.fnProgressUpdater = fnProgressUpdater;
  }

};

/**
 * Main object
 */
var Maf = {

  /**
   * Extract the archive using the specified program
   */
  extractFromArchive: function(program, archivefile, destpath) {
    /** If program is nothing then don't try to run it. */
    if (program != "") {

      if (MafPreferences.win_invisible) {
        localProgram = MafPreferences.win_wscriptexe;
        localProgramArgs = new Array();
        localProgramArgs[localProgramArgs.length] = MafPreferences.win_invisiblevbs;
        localProgramArgs[localProgramArgs.length] = program;
      } else {
        localProgram = program;
        localProgramArgs = new Array();
      }

      try {
        var oProgram = Components.classes[localFileContractID].getService(localFileIID);
        oProgram.initWithPath(localProgram);
      } catch(e) {
        alert("Could not find program: " + program + " \n" + e);
      }

      try {
        var oProcess = Components.classes[processContractID].createInstance(processIID);
      } catch (e) {
        alert("Could not create process:\n" + e);
      }

      oProcess.init(oProgram);

      localProgramArgs[localProgramArgs.length] = archivefile;
      localProgramArgs[localProgramArgs.length] = destpath;

      oProcess.run(true, localProgramArgs, localProgramArgs.length);
    }
  },

  /**
   * Save a single web page in an archive
   */
  saveAsWebPageComplete: function(aDocument, tempPath, scriptPath, archivePath) {
    var dateTimeArchived = new Date();

    var objMafArchiver = new MafArchiver(aDocument, tempPath, scriptPath, archivePath, dateTimeArchived);
    objMafArchiver.start();
  },

  /**
   * Save all open tabs in an archive
   */
  saveAllTabsComplete: function(browsers, tempPath, scriptPath, archivePath) {
    var objMafTabArchiver = new MafTabArchiver(browsers, tempPath, scriptPath, archivePath);
    objMafTabArchiver.start();
    MafUtils.showDownloadTabsDLG(objMafTabArchiver);
  },

  /**
   * Open a MAF archive and add the meta-data to the global state
   */
  openFromArchive: function(tempPath, scriptPath, archivePath) {
    var dateTimeExpanded = new Date();

    folderNumber = dateTimeExpanded.valueOf()+"_"+Math.floor(Math.random()*1000);

    var objMafTabExpander = new MafTabExpander(tempPath, scriptPath, archivePath, folderNumber);
    objMafTabExpander.start();
    MafUtils.showOpenTabsDLG(objMafTabExpander);

    MafState.openArchives.push(MafUtils.appendToDir(tempPath, folderNumber));
    var archiveLocalURLs = MafState.addArchiveInfo(tempPath, folderNumber, archivePath);

    if (MafPreferences.archiveOpenMode == MafPreferences.OPENMODE_ALLTABS) {
      MafUtils.openListInTabs(archiveLocalURLs);
    }

    if (MafPreferences.archiveOpenMode == MafPreferences.OPENMODE_SHOWDIALOG) {
      MafGUI.browseOpenArchives();
    }
  }

};


/**
 * Holds GUI related logic.
 */
var MafGUI = {

  /**
   * Prompts the user for the archive to load and then loads it.
   */
  loadArchive: function() {
    var archiveToOpen = this.selectFileOpen();

    Maf.openFromArchive(MafPreferences.temp,
                        MafPreferences.programFromOpenIndex(archiveToOpen[0]), archiveToOpen[1]);
  },

  /**
   * Prompts the user for the archive to save to and then adds the page to it.
   */
  addToArchive: function() {
    var archiveToAddTo = this.selectFileSave();

    Maf.saveAsWebPageComplete(window._content.document, MafPreferences.temp,
                              MafPreferences.programFromSaveIndex(archiveToAddTo[0]), archiveToAddTo[1]);
  },

  /**
   * Prompts the user for the archive to save to and then adds all tabbed pages to it.
   */
  addAllTabsToArchive: function() {
    var archiveToAddTo = this.selectFileSave();

    Maf.saveAllTabsComplete(window.getBrowser().browsers, MafPreferences.temp,
                            MafPreferences.programFromSaveIndex(archiveToAddTo[0]), archiveToAddTo[1]);
  },

  /**
   * Shows the Preferences
   */
  showPreferences: function() {
    MafUtils.showPreferencesDLG();
  },

  /**
   * Opens a File choose dialog with a save mode.
   * @return The file selected.
   */
  selectFileSave: function() {
    var filters = MafPreferences.getSaveFilters();

    var result = this.selectFile("Select MAF Archive:",
                                  filePickerIID.modeSave,
                                  filters);

    var selectedFileType = filters[result[1]][1];

    selectedFileType = selectedFileType.substring(1,selectedFileType.length);

    var filename = result[0].path;
    // if the file name does not end with the file type specified, tack it on

    if (filename.substring(filename.length-selectedFileType.length, filename.length).toLowerCase() !=
        selectedFileType.toLowerCase()) {
        filename += selectedFileType;
    }

    return [result[1], filename];
  },

  /**
   * Opens a File choose dialog with a open mode.
   * @return The file selected.
   */
  selectFileOpen: function() {
    var filters = MafPreferences.getOpenFilters();

    var result = this.selectFile("Select MAF Archive:",
                                  filePickerIID.modeOpen,
                                  filters);

    return [result[1], result[0].path];
  },

  /**
   * Shows the filepicker dialog with the appropriate filters.
   * @return The file selected.
   */
  selectFile: function(windowTitle, filePickerMode, filters) {
    var fp = Components.classes[filePickerContractID].createInstance(filePickerIID);
    fp.init(window, windowTitle, filePickerMode);


    for (i=0; i<filters.length; i++) {
      var title = filters[i][0];
      var mask = filters[i][1];
      fp.appendFilter(title, mask);
    }

    if (filters.length==0) {
      fp.appendFilters(nsIFilePicker.filterAll);
    }
    var res=fp.show();
    return [fp.file, fp.filterIndex];
  },


  /**
   * Shows the filepicker dialog with the directory select mode.
   * @return The directory selected as a string.
   */
  selectDirectory: function(windowTitle, initialDirectory) {
    var result = initialDirectory;

    var fp = Components.classes[filePickerContractID].createInstance(filePickerIID);
    fp.init(window, windowTitle, filePickerIID.modeGetFolder);

    try {
      if (initialDirectory != null) {
        // Create a directory reference to use
        var dir = Components.classes[localFileContractID].getService(localFileIID);
        dir.initWithPath(initialDirectory);

        fp.displayDirectory = dir;
      }
    } catch(e) {

    }

    var res = fp.show();

    if (res == Components.interfaces.nsIFilePicker.returnOK) {
      var selDir = fp.file.QueryInterface(localFileIID);
      result = selDir.path;
    }

    return result;
  },

  /**
   * Shows the file picker dialog allowing the user to select ONLY a specific file.
   * @return The full file path and filename as a string
   */
  selectSpecificFile: function(windowTitle, initialFilePath, filterFilename) {
    var result = initialFilePath;

    try {
      var filter = [ [filterFilename, filterFilename] ];
      var fresult = this.selectFile(windowTitle,
                                     filePickerIID.modeOpen,
                                     filter);
      result = fresult[0].path;
    } catch(e) {

    }

    return result;
  },

  /**
   *  Browse the open archives that are available.
   */
  browseOpenArchives: function() {
    MafUtils.showBrowseOpenArchivesDLG();
  },

  /**
   * Show the about dialog
   */
  showAbout: function() {
    MafUtils.showAboutDLG();
  }

};


/**
 * Holds the state of the MAF archives, the meta-data for each open archive.
 * This is a local object, so for each window this is created.
 * A global object stored in a hidden window holds the real state.
 * If that is null, that state is saved using this object.
 * If it is not null, this object is ignored.
 */
var MafState = {

  /** A list of open archives. */
  openArchives: new Array(),

  /** A count of the number of open archives. */
  noOfArchives: 0,

  /**
   * Add archive file info to the state.
   */
  addArchiveInfo: function(tempPath, foldernum, archivePath) {
    this.noOfArchives++;

    var archiveRootSubjectStr = "http://maf.mozdev.org/metadata/rdf/archive/" + this.noOfArchives + "/";

    archive1Subject = gRDFService.GetResource(archiveRootSubjectStr);
    archive1Sequence = gRDFCService.MakeSeq(this.datasource, archive1Subject);

    var localUrls = this.addArchivePagesToDatasource(tempPath, foldernum, archiveRootSubjectStr, archive1Sequence);

    var archiveRootSubject = gRDFService.GetResource(archiveRootSubjectStr);
    var archivePredicate = gRDFService.GetResource(MAFNamespace + "title");
    var archiveObject = gRDFService.GetResource("Archive " + this.noOfArchives);

    this.datasource.Assert(archiveRootSubject, archivePredicate, archiveObject, true);

    archivePredicate = gRDFService.GetResource(MAFNamespace + "localurl");
    archiveObject = gRDFService.GetResource(archivePath);

    this.datasource.Assert(archiveRootSubject, archivePredicate, archiveObject, true);

    this.archivesSequence.AppendElement(archive1Subject);

    return localUrls;
  },

  /**
   * For each saved page, there's an RDF file, process it to get the information.
   * Processing is split up into three loops because on windows local files initWithPath
   * causes something to freak out and stop working. Works fine as one loop in Linux.
   */
  addArchivePagesToDatasource: function(temp, expandedArchiveRoot, archiveSubjectRoot, archiveSequence) {
    try {

    var localUrls = new Array();

    var pageNo = 1;

    var oDir = Components.classes[localFileContractID].getService(localFileIID);
    oDir.initWithPath(MafUtils.appendToDir(temp, expandedArchiveRoot));

    if (oDir.exists() && oDir.isDirectory()) {
      var entries = oDir.directoryEntries;

      var dirList = new Array();

      // For each folder in the expanded archive root
      while (entries.hasMoreElements()) {

        var currDir = entries.getNext();
        currDir.QueryInterface(localFileIID);

        if (currDir.isDirectory()) {
          var currArchivePath = MafUtils.getURI(currDir.nsIFile);

          dirList[dirList.length] = [currArchivePath, currDir.path]

        }
      }


      var indexList = new Array();

      for (var i=0; i<dirList.length; i++) {

          var indexrdffile = Components.classes[localFileContractID].getService(localFileIID);
          indexrdffile.initWithPath(dirList[i][1]);
          indexrdffile.append("index.rdf");

          var title = "Page " + pageNo;
          var originalurl = "Unknown";
          var archivetime = "Unknown";
          var indexfilename = "index.html";

            // If the metadata exists
            if (indexrdffile.exists()) {
              // Update the variables with the actual metadata info
              var rdfdataresult = this.getMetaDataFrom(indexrdffile, dirList[i][0]);
              if (rdfdataresult["title"] != "") {
                title = rdfdataresult["title"];
              }
              if (rdfdataresult["originalurl"] != "") {
                originalurl = rdfdataresult["originalurl"];
              }
              if (rdfdataresult["archivetime"] != "") {
                archivetime = rdfdataresult["archivetime"];
              }
              if (rdfdataresult["indexfilename"] != "") {
                indexfilename = rdfdataresult["indexfilename"];
              }
            }


         indexList[indexList.length] = [title, originalurl, archivetime, indexfilename];
      }


      for (var i=0; i<dirList.length; i++) {
          var indexhtmlfile = Components.classes[localFileContractID].getService(localFileIID);
          indexhtmlfile.initWithPath(dirList[i][1]);
          indexhtmlfile.append(indexList[i][3]);

          var thisPageRDFSubjectRoot = archiveSubjectRoot + "" + (i+1) + "/";

          // get file URL - temp + expandedArchiveRoot + currFolder + index.html
          var localurl = MafUtils.getURI(indexhtmlfile.nsIFile);

          localUrls[localUrls.length] = localurl;

          this.addPageInfoToMetaData(thisPageRDFSubjectRoot, indexList[i][0], indexList[i][1], indexList[i][2], localurl, archiveSequence);
      }

    }

    } catch(e) {
      alert(e);
    }

    return localUrls;
  },

  /**
   * Asserts the RDF info.
   */
  addPageInfoToMetaData: function(thisPageRDFSubjectRoot, title, originalurl, archivetime, localurl, archiveSequence) {
    // Add the page info to the metadata

    // This archive's unique archive URL resource
    var rootSubject = gRDFService.GetResource(thisPageRDFSubjectRoot);

    // Add the title
    var predicate = gRDFService.GetResource(MAFNamespace + "title");
    var object = gRDFService.GetResource(title);
    this.datasource.Assert(rootSubject, predicate, object, true);

    // Add the original url
    predicate = gRDFService.GetResource(MAFNamespace + "originalurl");
    object = gRDFService.GetResource(originalurl);
    this.datasource.Assert(rootSubject, predicate, object, true);

    // Add the archive time
    predicate = gRDFService.GetResource(MAFNamespace + "archivetime");
    object = gRDFService.GetResource(archivetime);
    this.datasource.Assert(rootSubject, predicate, object, true);

    // Add the local url
    predicate = gRDFService.GetResource(MAFNamespace + "localurl");
    object = gRDFService.GetResource(localurl);
    this.datasource.Assert(rootSubject, predicate, object, true);

    archiveSequence.AppendElement(rootSubject);
  },

  /**
   * Tries to read the data from the RDF for a specific file.
   */
  getMetaDataFrom: function(sourcefile, resourcePath) {
    var result = new Array();
    result["title"] = "Unknown";
    result["originalurl"] = "Unknown";
    result["archivetime"] = "Unknown";
    result["indexfilename"] = "index.html";

    // Load a remote data source
    var mdatasource = Components.classes[xmlRDFDatasourceContractID].createInstance(xmlRDFDatasourceIID);
    mdatasource.Init(MafUtils.getURI(sourcefile.nsIFile));
    mdatasource.Refresh(true);
    mdatasource.QueryInterface(rdfDatasourceIID);

    try {
      // This archive's unique archive URL resource
      var rootSubject = gRDFService.GetResource("urn:root");

      // Get the title
      var predicate = gRDFService.GetResource(MAFNamespace + "title");

      var titletarget = mdatasource.GetTarget(rootSubject, predicate, true);
      titletarget = titletarget.QueryInterface(Components.interfaces.nsIRDFResource);
      result["title"] = titletarget.Value;
      result["title"] = result["title"].substring(resourcePath.length, result["title"].length);

      // Get the original url
      predicate = gRDFService.GetResource(MAFNamespace + "originalurl");

      var originalurltarget = mdatasource.GetTarget(rootSubject, predicate, true);
      originalurltarget = originalurltarget.QueryInterface(Components.interfaces.nsIRDFResource);
      result["originalurl"] = originalurltarget.Value;

      // Get the archive time
      predicate = gRDFService.GetResource(MAFNamespace + "archivetime");

      var originalurltarget = mdatasource.GetTarget(rootSubject, predicate, true);
      originalurltarget = originalurltarget.QueryInterface(Components.interfaces.nsIRDFResource);
      result["archivetime"] = originalurltarget.Value;
      result["archivetime"] = result["archivetime"].substring(resourcePath.length, result["archivetime"].length);

      // Get the index file name
      predicate = gRDFService.GetResource(MAFNamespace + "indexfilename");

      var originalurltarget = mdatasource.GetTarget(rootSubject, predicate, true);
      originalurltarget = originalurltarget.QueryInterface(Components.interfaces.nsIRDFResource);
      result["indexfilename"] = originalurltarget.Value;
      result["indexfilename"] = result["indexfilename"].substring(resourcePath.length, result["indexfilename"].length);

    } catch (e) {

    }

    return result;
  },


  setupDataSource: function() {
    var ds = Components.classes['@mozilla.org/rdf/datasource;1?name=in-memory-datasource'].createInstance();
    this.datasource = ds.QueryInterface(Components.interfaces.nsIRDFDataSource);

    var rootArchiveSubject = gRDFService.GetResource("http://maf.mozdev.org/metadata/rdf/open-archives");
    this.archivesSequence = gRDFCService.MakeSeq(this.datasource, rootArchiveSubject);

  }


};


/**
 * Helper function that lets one see the source to in memory rdf data sources
 */
function serialize(originalDatasource){
  var ser = Components.classes["@mozilla.org/rdf/xml-serializer;1"].getService(Components.interfaces.nsIRDFXMLSerializer);

  ser.QueryInterface(Components.interfaces.nsIRDFXMLSource);

  var outputstream = {
    content:"",
    write:function(s, count) {
      this.content+=s;
      return count;
    },
    flush:function(){},
    close:function(){}
  };

  ser.init(originalDatasource);
  ser.Serialize(outputstream);
  alert(outputstream.content);
};


/**
 * Utility functions. Some functionality may still be moved here to increase modularity.
 */
var MafUtils = {

  /**
   * Cross platform append
   */
  appendToDir: function(initialDirectory, subDirectory) {
    var result = initialDirectory;
    try {
      var dir = Components.classes[localFileContractID].getService(localFileIID);
      dir.initWithPath(initialDirectory);
      dir.append(subDirectory);
      result = dir.path;
    } catch(e) {

    }
    return result;
  },

  /**
   * Create RDF file based on template.
   */
  createRDF: function(path, filename) {
    var dir = Components.classes[localFileContractID].getService(localFileIID);
    dir.initWithPath(path);
    dir.append(filename);

    try {
      var oFile = Components.classes[localFileContractID].createInstance(localFileIID);
      oFile.initWithPath(dir.path);
      if (!oFile.exists()) {
        oFile.create(0x00, 0644);
      }
    } catch (e) {
      alert(e);
    }

    try {
      var oTransport = Components.classes[fileOutputStreamContractID].createInstance(fileOutputStreamIID);
      oTransport.init( oFile, 0x04 | 0x08 | 0x10, 064, 0 );
      oTransport.write(MAFRDFTemplate, MAFRDFTemplate.length);
      oTransport.close();
    } catch (e) {
      alert(e);
    }

    // Load a remote data source
    var datasource = Components.classes[xmlRDFDatasourceContractID].createInstance(xmlRDFDatasourceIID);
    datasource.Init(MafUtils.getURI(oFile.nsIFile));
    datasource.Refresh(true);

    return datasource;
  },

  /**
   * Get the URL of the local file specified.
   */
  getURI: function(nsIFile) {
    var serv = Components.classes[ioServiceNetworkContractID].getService(ioServiceNetworkIID);
    var uri = serv.newFileURI(nsIFile);
    return uri.spec;
  },

  /**
   * Add string data to the data source.
   */
  addStringData: function(datasource, name, value) {
    var rootSubject = gRDFService.GetResource("urn:root");
    var predicate = gRDFService.GetResource(MAFNamespace + name);
    var object = gRDFService.GetResource(value);

    // Make sure we have an interface that we can assert to
    modDataSource = datasource.QueryInterface(rdfDatasourceIID);

    modDataSource.Assert(rootSubject,predicate,object,true);
  },

  /**
   * Shows the dialog to the user when saving tabs.
   */
  showDownloadTabsDLG: function(objMafArchiver) {
    const url = "chrome://maf/content/mafSaveTabsDLG.xul";

    var w = 400;
    var h = 50;

    var sX = (window.screen.width/2)-(Math.round(w/2));
    var sY = (window.screen.height/2)-(Math.round(h/2));

    var win_prefs = "chrome,dialog,dependent=no,modal,resizable=yes,screenX="+ sX + ",screenY="+ sY +
                    ",width="+ w +",height=" + h;
    window.openDialog(url, "_blank", win_prefs, objMafArchiver);
  },

  /**
   * Shows the dialog to the user when extracting files.
   */
  showOpenTabsDLG: function(objMafExpander) {
    const url = "chrome://maf/content/mafOpenTabsDLG.xul";

    var w = 400;
    var h = 50;

    var sX = (window.screen.width/2)-(Math.round(w/2));
    var sY = (window.screen.height/2)-(Math.round(h/2));

    var win_prefs = "chrome,dialog,dependent=no,modal,resizable=yes,screenX="+ sX + ",screenY="+ sY +
                    ",width="+ w +",height=" + h;
    window.openDialog(url, "_blank", win_prefs, objMafExpander);
  },

  /**
   * Shows the user some info about MAF
   */
  showAboutDLG: function() {
    const url = "chrome://maf/content/mafAboutDLG.xul";

    var win_prefs = "chrome,dialog,dependent=no,modal,resizable=yes";

    window.openDialog(url, "_blank", win_prefs);
  },

  /**
   * Shows the user the preferences dialog
   */
  showPreferencesDLG: function() {
    const url = "chrome://maf/content/mafPreferencesDLG.xul";

    var w = 500;
    var h = 500;

    var sX = (window.screen.width/2)-(Math.round(w/2));
    var sY = (window.screen.height/2)-(Math.round(h/2));

    var win_prefs = "chrome,dialog,dependent=no,modal,resizable=yes,screenX="+ sX + ",screenY="+ sY +
                    ",width="+ w +",height=" + h;
    window.openDialog(url, "_blank", win_prefs, MafPreferences, MafGUI);
  },

  /**
   * Shows the user a window that allows them to browse the open archives.
   */
  showBrowseOpenArchivesDLG: function() {
    const url = "chrome://maf/content/mafBrowseOpenArchivesDLG.xul";

    var w = 500;
    var h = 500;

    var sX = (window.screen.width/2)-(Math.round(w/2));
    var sY = (window.screen.height/2)-(Math.round(h/2));

    var win_prefs = "chrome,dialog,dependent=no,modal,resizable=yes,screenX="+ sX + ",screenY="+ sY +
                    ",width="+ w +",height=" + h;
    window.openDialog(url, "_blank", win_prefs, MafState, window);
  },

  /**
   * Obsolete method that attempts to open all expanded folders in a directory in tabs.
   * Does not use meta-data to determine what the index file is.
   */
  openDirInTabs: function(temp, expandedArchiveRoot) {
    var oBrowser = window.getBrowser();

    try {

    var oDir = Components.classes[localFileContractID].getService(localFileIID);
    oDir.initWithPath(MafUtils.appendToDir(temp, expandedArchiveRoot));

    if (oDir.exists() && oDir.isDirectory()) {

      var entries = oDir.directoryEntries;

      // For each folder in the expanded archive root
      while (entries.hasMoreElements()) {
        currDir = entries.getNext();
        currDir.QueryInterface(localFileIID);

        if (currDir.isDirectory()) {

          currDir.append("index.html");

          // If index.html exists
          if (currDir.exists()) {
            // get file URL - temp + expandedArchiveRoot + currFolder + index.html
            url = this.getURI(currDir.nsIFile);
            oBrowser.addTab(url);
          }

        }

      }
    }

    } catch(e) {
      alert(e);
    }

  },

  /**
   * Opens a list of URLs in tabs.
   */
  openListInTabs: function(urlList) {
    var oBrowser = window.getBrowser();

    try {
      var triedFirstTab = false;
      for (i=0; i<urlList.length; i++) {
        if (triedFirstTab) {
          oBrowser.addTab(urlList[i]);
        } else {
          triedFirstTab = true;
          if ((oBrowser.browsers.length == 1) && (oBrowser.currentURI.spec == "about:blank")) {
            oBrowser.loadURI(urlList[i], null, null);
          } else {
            oBrowser.addTab(urlList[i]);
          }

        }
      }
    } catch(e) {

    }

  },

  /**
   * Returns the number of open windows
   */
  getNumberOfOpenWindows: function() {
    var numberOfOpenWindows = 0;

    try {
      var wmI = Components.classes[asWinMedContractID].getService(asWinMedIID);
      var entries = wmI.getEnumerator(null);

      while (entries.hasMoreElements()) {
        currWindow = entries.getNext();
        numberOfOpenWindows++;
      }
    } catch (e) {
      alert(e);
    }

    return numberOfOpenWindows;
  },

  /**
   * Registered trigger event whenever a window is closed.
   */
  onWindowClose: function(evt) {

    // Check to see it's the last open window
    var numberOfOpenWindows = MafUtils.getNumberOfOpenWindows();

    // If it's the last window
    if (numberOfOpenWindows < 2) {
      // Get the temp folders used by open archive functions
      // delete each of those folders

      for (i=0; i<MafState.openArchives.length; i++) {
        try {
          var oDir = Components.classes[localFileContractID].getService(localFileIID);
          oDir.initWithPath(MafState.openArchives[i]);

          if (oDir.exists() && oDir.isDirectory()) {
            oDir.remove(true);
          }
        } catch(e) {

        }
      }

      try {
        // The code below is technically not necessary, as this should only happen
        // when there are no more windows (app is closed). Data in the hidden window
        // should be gone too.

        // Get hidden window
        var appShell = Components.classes[appShellContractID].getService(appShellIID);
        hiddenWnd = appShell.hiddenDOMWindow;
        hiddenWnd.MafState = null;
      } catch(e) {

      }
    }
  },

  /**
   * Whenever a new window is opened.
   */
  onWindowLoad: function(evt) {
     // Get hidden window
     var appShell = Components.classes[appShellContractID].getService(appShellIID);
     hiddenWnd = appShell.hiddenDOMWindow;

     // Store global MafState in hidden window
     if (typeof(hiddenWnd.MafState) == "undefined") {
       hiddenWnd.MafState = MafState;
     } else {
       MafState = hiddenWnd.MafState;
     }
  }

};

/**
 * The MAF preferences.
 */
var MafPreferences = {

  /** The temp folder. */
  temp: "",

  /** The registered scripts used with archiving. */
  programExtensions: new Array(),

  /** The extension that handles *.maf by default. */
  defaultMAFExtensionIndex: 0,

  archiveOpenMode: 0,
     /** 0 - Do nothing. */
     /** 1 - Open all in new tabs. */
     /** 2 - Dialog box showing all archived files, select to open. */

  OPENMODE_NOTHING: 0,

  OPENMODE_ALLTABS: 1,

  OPENMODE_SHOWDIALOG: 2,

  /** URL Rewrite enabled. */
  urlRewrite: false,

  /** Save extended metadata. */
  saveExtendedMetadata: true,

  /** Windows specific preferences. */
  win_invisible: false,

  win_wscriptexe: "c:\\winnt\\system32\\wscript.exe",

  win_invisiblevbs: "c:\\temp\\invis.vbs",

  /**
   * Creates a multi-dimensional array holding info on each registered program
   */
  getOpenFilters: function() {
    var result = [ ["MAF Archives", "*.maf", this.defaultMAFExtensionIndex] ];
    for (i=0; i<this.programExtensions.length; i++) {
      var entry = ["MAF " + this.programExtensions[i][0] + " Archives"];

      // Construct a string like "*.zip.maf; *.maf.zip"
      var additionalExts = "";
      for (j=0; j<this.programExtensions[i][3].length; j++) {
        if (additionalExts == "") {
          additionalExts = this.programExtensions[i][3][j];
        } else {
          additionalExts += "; " + this.programExtensions[i][3][j];
        }
      }
      entry[entry.length] = additionalExts;
      // Add associated program extension index
      entry[entry.length] = i;
      result[result.length] = entry;
    }
    return result;
  },

  /**
   * Creates a multi-dimensional array holding info on each registered program
   */
  getSaveFilters: function() {
    var result = [ ["MAF Archives", "*.maf", this.defaultMAFExtensionIndex] ];

    // Each unique extension has its own entry
    for (i=0; i<this.programExtensions.length; i++) {
      for (j=0; j<this.programExtensions[i][3].length; j++) {
        var entry = ["MAF " + this.programExtensions[i][0] + " Archives", this.programExtensions[i][3][j], i];
        result[result.length] = entry;
      }
    }
    return result;
  },

  /**
   * Gets the program to use from the selected index.
   */
  programFromSaveIndex: function(index) {
    filters = this.getSaveFilters();
    selProgExt = this.programExtensions[filters[index][2]];
    return selProgExt[1];
  },

  /**
   * Gets the program to use from the selected index.
   */
  programFromOpenIndex: function(index) {
    filters = this.getOpenFilters();
    selProgExt = this.programExtensions[filters[index][2]];
    return selProgExt[2];
  },

  /**
   * Load the preferences from the user prefs.
   */
  load: function() {
    // Default if there's no stored prefs
    this.temp = "/tmp/maftmp/";

    this.defaultMAFExtensionIndex = 0;

    this.programExtensions[this.programExtensions.length] = [
       "Zip", "/tmp/mafzip.sh", "/tmp/mafunzip.sh", ["*.zip.maf", "*.maf.zip"]];

    // Load the temp path
    // Load the defaultMAFExtensionIndex
    // Load the program extensions array
      // For each extension
        // Load the extension's id
        // Load the archive program path + filename
        // Load the extract program path + filename
        // Load each of the extensions wildcard file matches
      // end-for

    try {

      var prefs = Components.classes[prefSvcContractID].getService(prefSvcIID).getBranch("maf.");

      this.temp = prefs.getCharPref("temp");
      this.urlRewrite = prefs.getBoolPref("urlrewrite");
      this.saveExtendedMetadata = prefs.getBoolPref("saveextendedmetadata");
      this.defaultMAFExtensionIndex = prefs.getIntPref("defaultmafhandler");
      this.archiveOpenMode = prefs.getIntPref("archiveopenmode");

      this.win_invisible = prefs.getBoolPref("wininvisible");
      this.win_wscriptexe = prefs.getCharPref("winwscriptexe");
      this.win_invisiblevbs = prefs.getCharPref("wininvisiblevbs");

      var noOfExtensions = prefs.getIntPref("noofextensions");

      this.programExtensions = new Array();

      for (i=0; i<noOfExtensions; i++) {
        currEntry = new Array();
        currEntry[0] = prefs.getCharPref("ext." + i + ".id");
        currEntry[1] = prefs.getCharPref("ext." + i + ".archive");
        currEntry[2] = prefs.getCharPref("ext." + i + ".extract");
        var noOfMasks = prefs.getIntPref("ext." + i + ".masklength");
        maskEntry = new Array();
        for (j=0; j<noOfMasks; j++) {
          maskEntry[j] = prefs.getCharPref("ext." + i + ".mask." + j);
        }
        currEntry[3] = maskEntry;
        this.programExtensions[this.programExtensions.length] = currEntry;
      }

    } catch(e) {
      // alert(e);
    }
  },

  /**
   * Save the preferences to the user prefs.
   */
  save: function() {
    // Save the temp path
    // Save the defaultMAFExtensionIndex
    // Save the program extensions array
      // For each extension
        // Save the extension's 3 letter id
        // Save the archive program path + filename
        // Save the extract program path + filename
        // Save each of the extensions wildcard file matches
      // end-for
    try {

      var prefs = Components.classes[prefSvcContractID].getService(prefSvcIID).getBranch("maf.");

      prefs.setCharPref("temp", this.temp);
      prefs.setBoolPref("urlrewrite", this.urlRewrite);
      prefs.setBoolPref("saveextendedmetadata", this.saveExtendedMetadata);
      prefs.setIntPref("defaultmafhandler", this.defaultMAFExtensionIndex);
      prefs.setIntPref("archiveopenmode", this.archiveOpenMode);

      prefs.setBoolPref("wininvisible", this.win_invisible);
      prefs.setCharPref("winwscriptexe", this.win_wscriptexe);
      prefs.setCharPref("wininvisiblevbs", this.win_invisiblevbs);

      prefs.setIntPref("noofextensions", this.programExtensions.length);
      for (i=0; i<this.programExtensions.length; i++) {
        prefs.setCharPref("ext." + i + ".id", this.programExtensions[i][0]);
        prefs.setCharPref("ext." + i + ".archive", this.programExtensions[i][1]);
        prefs.setCharPref("ext." + i + ".extract", this.programExtensions[i][2]);
        prefs.setIntPref("ext." + i + ".masklength", this.programExtensions[i][3].length);
        for (j=0; j<this.programExtensions[i][3].length; j++) {
          prefs.setCharPref("ext." + i + ".mask." + j, this.programExtensions[i][3][j]);
        }
      }

    } catch(e) {
      alert(e);
    }
  }

};


// Setup the RDF in memory data source for the current state
MafState.setupDataSource();

// Load the last stored preferences into the object.
MafPreferences.load();

window.addEventListener("close", MafUtils.onWindowClose, true);
window.addEventListener("load", MafUtils.onWindowLoad, true);
